/*
    libporty - async background processes for UNIX
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

void P_net_async_bgproc_closefds(P_net_socket *inherit_list, int inherit_len)
{
	int n, max;
	char *inherit;

	max = 0;
	for(n = 0; n < inherit_len; n++) {
		if (inherit_list[n] > max)
			max = inherit_list[n];
	}
	max++;
	inherit = calloc(max, 1);

	for(n = 0; n < inherit_len; n++)
		inherit[inherit_list[n]] = 1;

	for(n = 3; n < 1024; n++) {
		if ((n < max) && (inherit[n]))
			continue;
		close(n);
	}

	free(inherit);
}

int P_net_async_bgproc_inherit(P_net_socket *fd, char *argv[], P_net_socket *ctrlfd, P_net_socket *fd_err, P_net_socket *inherit_list, int inherit_len)
{
	pid_t pid;
	P_net_socket chfd, chfd_err;
	int ret;

	ret = P_net_local_socket(fd, &chfd, P_net_nonblock_full);
	if (ret != 0)
		return ret;
	if (fd_err != NULL) {
		ret = P_net_local_socket(fd_err, &chfd_err, P_net_nonblock_full);
		if (ret != 0)
			return ret;
	}

	children_created++;
	pid = fork();
	if (pid < 0)
		return -128;
	if (pid == 0) {
		/* rebind stdio to the child side of the socket pair */
		P_net_set_blocking(chfd);
		close(0);
		dup(chfd);
		close(1);
		dup(chfd);
		if (fd_err != NULL) {
			close(2);
			dup(chfd_err);
		}
		P_net_async_bgproc_closefds(inherit_list, inherit_len);
		execvp(argv[0], argv);
		exit(1);
	}
	P_net_close(chfd);

	if (ctrlfd != NULL) {
		int cfd;
		P_net_socket inv = P_NET_INVALID_SOCKET;
		ret = P_net_local_socket(ctrlfd, &cfd, P_net_nonblock_full);
		if (ret != 0)
			return ret;

		watchdog_add(&cfd, &pid, &inv);
	}
	return 0;
}

int P_net_async_bgproc(P_net_socket *fd, char *argv[], P_net_socket *ctrlfd)
{
	return P_net_async_bgproc_inherit(fd, argv, ctrlfd, NULL, NULL, 0);
}

static int watchdog_wait(P_net_socket cfd, int numpids, P_pid_t *pids, bgrpoc_status_t *exit_status)
{
	pid_t pid;
	int status;
	int n;

	/* check what's happened */
	pid = waitpid(-1, &status, WNOHANG);

/*	printf("Wait return: %d\n", pid);*/

	/* look up pid in the database */
	for(n = 0; n < numpids; n++) {
		if (pids[n] == pid) {
			/* found the process */
			if (WIFEXITED(status)) {
				exit_status->reason = BGPROC_EXITED;
				exit_status->code   = WEXITSTATUS(status);
			}
			else  if (WIFSIGNALED(status) && ((WTERMSIG(status) == 15) || (WTERMSIG(status) == 9))) {
				exit_status->reason = BGPROC_KILLED;
				exit_status->code   = 0xFFFF;
			}
			else {
				exit_status->reason = BGPROC_ABNORMAL;
				exit_status->code   = 0xFFFF;
			}
			return n;
		}
	}

	/* not our child */
	return -1;
}

static void watchdog_kill(P_pid_t pid, int hard)
{
	kill(pid, hard ? 9 : 15);
}

