/*
 * Copyright (c) 2003 Anders Magnusson (ragge@ludd.luth.se).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright(C) Caldera International Inc. 2001-2002. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code and documentation must retain the above
 * copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * All advertising materials mentioning features or use of this software
 * must display the following acknowledgement:
 *         This product includes software developed or owned by Caldera
 *        International, Inc.
 * Neither the name of Caldera International, Inc. nor the names of other
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * USE OF THE SOFTWARE PROVIDED FOR UNDER THIS LICENSE BY CALDERA
 * INTERNATIONAL, INC. AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL CALDERA INTERNATIONAL, INC. BE LIABLE
 * FOR ANY DIRECT, INDIRECT INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OFLIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

 /***************************************************************************
    This is a modified verison for c99tree; these modifications are
    Copyright (c) 2010 Tibor Palinkas.
    Original file version marking:
    cgram.y,v 1.296 2010/08/11 14:08:44 ragge Exp $
    Modifications are licensed the same way as the original file.
 *****************************************************************************/

/*
 * Comments for this grammar file. Ragge 021123
 *
 * ANSI support required rewrite of the function header and declaration
 * rules almost totally.
 *
 * The lex/yacc shared keywords are now split from the keywords used
 * in the rest of the compiler, to simplify use of other frontends.
 */

/*
 * At last count, there were 3 shift/reduce and no reduce/reduce conflicts
 * Two was funct_idn and the third was "dangling else".
 * (currently we have some more -- Igor2)
 */

/*%locations*/
%error-verbose


/*
 * Token used in C lex/yacc communications.
 */
%token        C_STRING        /* a string constant */
%token        C_ICON          /* an integer constant */
%token        C_FCON          /* a floating point constant */
%token        C_NAME          /* an identifier */
%token        C_TYPENAME      /* a typedef'd name */
%token        C_ANDAND        /* && */
%token        C_OROR          /* || */
%token        C_GOTO          /* unconditional goto */
%token        C_RETURN        /* return from function */
%token        C_TYPE          /* a type */
%token        C_TYPESPEC      /* a type specifier (signed/unsigned) */
%token        C_CLASS         /* a storage class */
%token        C_TYPEDEF       /* typedef command */
%token        C_ASOP          /* assignment ops */
%token        C_RELOP         /* <=, <, >=, > */
%token        C_EQUOP         /* ==, != */
%token        C_DIVOP         /* /, % */
%token        C_SHIFTOP       /* <<, >> */
%token        C_INCOP         /* ++, -- */
%token        C_UNOP          /* !, ~ */
%token        C_STROP         /* ., -> */
%token        C_STRUCT
%token        C_IF
%token        C_ELSE
%token        C_SWITCH
%token        C_BREAK
%token        C_CONTINUE
%token        C_WHILE
%token        C_DO
%token        C_FOR
%token        C_DEFAULT
%token        C_CASE
%token        C_SIZEOF
%token        C_ALIGNOF
%token        C_ENUM
%token        C_ELLIPSIS
%token        C_QUALIFIER
%token        C_INLINE
%token        C_ASM
%token        NOMATCH
%token        C_TYPEOF        /* COMPAT_GCC */
%token        C_ATTRIBUTE     /* COMPAT_GCC */
%token        PCC_OFFSETOF
%token        C_PRAGMA
%token        C_COMMENT

/*
 * Precedence
 */
%left ','
%right '=' C_ASOP
%right '?' ':'
%left C_OROR
%left C_ANDAND
%left '|'
%left '^'
%left '&'
%left C_EQUOP
%left C_RELOP
%left C_SHIFTOP
%left '+' '-'
%left '*' C_DIVOP
%right C_UNOP
%right C_INCOP C_SIZEOF
%left '[' '(' C_STROP
%{
# include "c99tree.h"
# include "tree.h"
# include "scope.h"
# include <stdarg.h>
# include <string.h>
# include <stdlib.h>
char *c_typename = NULL;
static const char *extract_id_from_cast(node_t *n);
static int is_designator(node_t *n);
extern node_t *AST;
%}

%union {
	node_t *node;
	char *string;
}

/* define types */
%start ext_def_list

%type <node>    external_def kr_args declaration asmstatement arg_dcl_list
                arg_declaration block_item_list block_item compoundstmt
                init_declarator init_declarator_list init_list label moe
                moe_list asm_mvol statement struct_dcl_list struct_declaration
                struct_declarator struct_declarator_list switch_label
                xnfdeclarator_typename init_declarator_typename

%type <node>    ifelprefix ifprefix whprefix forprefix doprefix switchpart

%type <node>    e .e ae be term enum_dcl struct_dcl cast_type
                declarator declarator_typename elist type_sq cf_spec
                merge_attribs
                parameter_declaration abstract_declarator initializer
                parameter_type_list parameter_list addrlbl
                declaration_specifiers designation type_sq_list
                identifier_list arg_param_list type_qualifier_list
                designator_list designator xasm oplist oper cnstr funtype
                typeof attribute attribute_specifier
                attribute_list attr_spec_list attr_var preproc
                abstract_declarator_ end  endc virt_statement

%type <node>    string string_
%type <string>  C_STRING
%type <node>    str_head
%type <node>    xnfdeclarator enum_head

%type <string>  C_STRUCT C_RELOP C_DIVOP C_SHIFTOP
                C_ANDAND C_OROR C_STROP C_INCOP C_UNOP C_ASOP C_EQUOP C_CASE

%type <string>  C_TYPESPEC C_TYPE C_QUALIFIER C_ICON C_FCON C_CLASS C_TYPEDEF
%type <node>    C_TYPENAME
%type <string>  C_NAME C_CONTINUE C_ELLIPSIS
%type <string>  C_PRAGMA C_COMMENT C_FOR C_WHILE C_GOTO C_RETURN C_TYPEOF
%type <string>  str_name C_SWITCH C_IF C_DO C_INLINE C_BREAK C_DEFAULT
%%

/* entry point - a list of external defs */
ext_def_list:
		  ext_def_list external_def         { if ($2 != NULL) {if (AST == NULL) AST = $2; else Nappend(AST, $2);} }
		| /* empty */
		;

/* source is made of declaration, function definition, preproc, or asm */
external_def:
		  funtype kr_args compoundstmt {
			$$ = Nalloc(T_FUNCDEF, strdup($1->next->name), $1);
			if (($1->next->children->name != NULL) && (strcmp($1->next->children->name, "K&R/1") == 0))
				Nchappend($1->next, Nalloc(T_ARGLIST, strdup("K&R/2"), $2));
			Nchappend($$, $3);
		}
		| declaration        { $$ = $1; }
		| asmstatement ';'   { $$ = Nalloc(T_FUNCDEF, NULL, $1); }
		| preproc            { $$ = $1; }
		| ';'                { $$ = NULL; }
		| error              { $$ = NULL; }
		;

/* function declaration */
funtype:
		  /* no type given */ declarator {
			$$ = Nalloc(T_TYPE, NULL, NULL);
			Nappend($$, $1);
		}
		| declaration_specifiers declarator {
			$$ = $1;
			Nappend($1, $2);
		}
		;

kr_args:
		  /* empty */       { $$ = NULL; }
		| arg_dcl_list      { $$ = $1;   }
		;

declaration_specifiers:
		  merge_attribs    { $$ = Nalloc(T_TYPE, NULL, $1); }
		;

merge_attribs:
		  type_sq                 { $$ = $1; }
		| merge_attribs type_sq   { $$ = Nappend($1, $2); }
		| cf_spec                 { $$ = $1; }
		| merge_attribs cf_spec   { $$ = Nappend($1, $2); }
		;

type_sq:
		  C_TYPE                { $$ = Nalloc(T_TYPENAME, $1, NULL); }
		| C_TYPESPEC            { $$ = Nalloc(T_TYPESPEC, $1, NULL); }
		| C_TYPENAME            { $$ = $1; free(c_typename); c_typename = NULL; }
		| struct_dcl            { $$ = $1; }
		| enum_dcl              { $$ = $1; }
		| C_QUALIFIER           { $$ = Nalloc(T_QUALIFIER, $1, NULL);  }
		| typeof                { $$ = $1; }
		| attribute_specifier   { $$ = $1; }
		;

cf_spec:
		  C_CLASS       { $$ = Nalloc(T_CLASS, $1, NULL); }
		| C_INLINE      { $$ = Nalloc(T_CLASS, $1, NULL); }
		;

typeof:
		  C_TYPEOF '(' e ')'            { $$ = Nalloc(T_TYPEOF, NULL, $3); }
		| C_TYPEOF '(' cast_type ')'    { $$ = Nalloc(T_TYPEOF, NULL, $3); }

		;

attribute_specifier :
		  C_ATTRIBUTE '(' '(' attribute_list ')' ')'   { $$ = Nalloc(T_ATTRLIST, NULL, $4); }
		;

attribute_list:
		  attribute                       { $$ = $1; }
		| attribute_list ',' attribute    { $$ = Nappend($1, $3); }
		;

attribute:
		  /* EMPTY */               { $$ = NULL; }
		| C_NAME                    { $$ = Nalloc(T_ATTR, $1, NULL); }
		| C_NAME '(' elist ')'      { $$ = Nalloc(T_ATTR, $1, $3); }
		| C_QUALIFIER               { $$ = Nalloc(T_ATTR, $1, NULL); }
		;

declarator:
		  '*' declarator                            { $$ = $2; NTYinsert($$, Nalloc(T_PTR, NULL, NULL)); }
		| '*' type_qualifier_list declarator        { $$ = $3; NTYinsert($$, $2); Nchinsert($$, Nalloc(T_PTR, NULL, NULL));}
		| C_NAME                                    { $$ = Nalloc(T_ID, $1, NULL); }
		| '(' attr_spec_list declarator ')'         { $$ = Nappend($2, $3); }
		| '(' declarator ')'                        { $$ = $2; }
		| declarator '[' ae ']'                     { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, NULL, $3)); }
		| declarator '[' merge_attribs  ae ']'      { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, NULL, $3)); Nappend($3, $4);}
		| declarator '[' merge_attribs ']'          { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, NULL, $3)); }
		| declarator '[' ']'                        { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, NULL, NULL)); }
		| declarator '[' '*' ']'                    { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, strdup("abstract_*"), Nalloc(T_ID, strdup("*"), NULL))); }
		| declarator '(' parameter_type_list ')'    { $$ = $1; NTYinsert($$, Nalloc(T_ARGLIST, NULL, $3)); }
		| declarator '(' identifier_list ')'        { $$ = $1; NTYinsert($$, Nalloc(T_ARGLIST, strdup("K&R/1"), $3)); }
		| declarator '(' ')'                        { $$ = $1; NTYinsert($$, Nalloc(T_ARGLIST, NULL, NULL)); }
		;

declarator_typename:
		  '*' declarator_typename                            { $$ = $2; NTYinsert($$, Nalloc(T_PTR, NULL, NULL)); }
		| '*' type_qualifier_list declarator_typename        { $$ = $3; NTYinsert($$, $2); Nchinsert($$, Nalloc(T_PTR, NULL, NULL));}
		| C_TYPENAME                                         { $$ = Nalloc(T_ID, strdup($1->name), NULL); scope_negative_typedef_add($1->name); Nfree($1); free(c_typename); c_typename = NULL; }
		| '(' attr_spec_list declarator_typename ')'         { $$ = Nappend($2, $3); }
		| '(' declarator_typename ')'                        { $$ = $2; }
		| declarator_typename '[' ae ']'                     { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, NULL, $3)); }
		| declarator_typename '[' ']'                        { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, NULL, NULL)); }
		| declarator_typename '[' '*' ']'                    { $$ = $1; NTYinsert($$, Nalloc(T_INDEX, strdup("*"), NULL)); }
		| declarator_typename '(' parameter_type_list ')'    { $$ = $1; NTYinsert($$, Nalloc(T_ARGLIST, NULL, $3)); }
		| declarator_typename '(' identifier_list ')'        { $$ = $1; NTYinsert($$, Nalloc(T_ARGLIST, strdup("K&R/1"), $3)); }
		| declarator_typename '(' ')'                        { $$ = $1; NTYinsert($$, Nalloc(T_ARGLIST, NULL, NULL)); }
		;

/* add qualifiers to a declaration */
type_qualifier_list:
		  C_QUALIFIER                                  { $$ = Nalloc(T_QUALIFIER, $1, NULL); }
		| type_qualifier_list C_QUALIFIER              { $$ = Nappend($1, Nalloc(T_QUALIFIER, $2, NULL)); }
		| attribute_specifier                          { $$ = $1; }
		| type_qualifier_list attribute_specifier      { $$ = Nappend($1, $2); }
		;

/* when a list of IDs is given without types - currently used only in K&R first set of arguments */
identifier_list:
		  C_NAME                          { $$ = Nalloc(T_ARG, NULL, Nalloc(T_ID, $1, NULL)); }
		| identifier_list ',' C_NAME      { $$ = Nappend($1, Nalloc(T_ARG, NULL, Nalloc(T_ID, $3, NULL))); }
		;

parameter_type_list:
		  parameter_list                    { $$ = $1; }
		| parameter_list ',' C_ELLIPSIS     { $$ = Nappend($1, Nalloc(T_ARG, NULL, Nalloc(T_ELLIPSIS, NULL, NULL))); }
		;

parameter_list:
		  parameter_declaration                       { $$ = Nalloc(T_ARG, NULL, $1); }
		| parameter_list ',' parameter_declaration    { $$ = Nappend($1, Nalloc(T_ARG, NULL, $3)); }
		;

parameter_declaration:
		  declaration_specifiers declarator attr_var {
			$$ = $1;
			Nappend($$, $2);
			Nappend($$, $3);
		}
		| declaration_specifiers abstract_declarator    { $$ = Nappend($1, $2); }
		| declaration_specifiers                        { $$ = $1; }
		;

/* anonymous declarations */
abstract_declarator:
	abstract_declarator_ { $$ = Nalloc(T_ID, strdup("<anonymous>"), $1); }
	;

abstract_declarator_:
		  '*'                             { $$ = Nalloc(T_PTR, NULL, NULL); }
		| '*' type_qualifier_list         { $$ = Nappend(Nalloc(T_PTR, NULL, NULL), $2); }
		| '*' abstract_declarator_        { $$ = Nappend(Nalloc(T_PTR, NULL, NULL), $2); }
		| '*' type_qualifier_list abstract_declarator_ {
			$$ = $3;
			Nappend($$, $2);
			Nappend($$, Nalloc(T_PTR, NULL, NULL));
		}
		| '(' abstract_declarator_ ')'   { $$ = $2; }
		| '[' ']' attr_var               { $$ = Nappend(Nalloc(T_INDEX, strdup("abstract_1"), NULL), $3); }
		| '[' ae ']' attr_var {
			$$ = Nalloc(T_INDEX, strdup("abstract_2"), NULL);
			Nchinsert($$, $2);
			Nappend($$, $4);
		}
		| abstract_declarator_ '[' ']' attr_var {
			if ($4 != NULL) {
				$$ = $4;
				Nappend($$, Nalloc(T_INDEX, strdup("abstract_3a"), NULL));
			}
			else
				$$ = Nalloc(T_INDEX, strdup("abstract_3b"), NULL);
			Nappend($$, $1);
		}
		| abstract_declarator_ '[' ae ']' attr_var {
			$$ = Nalloc(T_INDEX, strdup("abstract_4"), $3);
			Nappend($$, $5);
			Nappend($$, $1);
		}
		| '(' ')'                           { $$ = Nalloc(T_ARGLIST, NULL, NULL); }
		| '(' parameter_type_list ')'       { $$ = Nalloc(T_ARGLIST, NULL, $2); }
		|  abstract_declarator_ '(' ')'     { $$ = Nappend(Nalloc(T_TYPE, NULL, Nalloc(T_ARGLIST, NULL, NULL)), $1); }
		|  abstract_declarator_ '(' parameter_type_list ')' { $$ = Nappend(Nalloc(T_TYPE, NULL, Nalloc(T_ARGLIST, NULL, $3)), $1); }
		;

/*
 * K&R arg declaration, between ) and {
 */
arg_dcl_list:
		  arg_declaration                   { $$ = $1; }
		| arg_dcl_list arg_declaration      { $$ = Nappend($1, $2); }
		;


arg_declaration:
		declaration_specifiers arg_param_list ';'  { $$ = Nalloc(T_ARG, NULL, Nappend($1, $2)); }
		;

arg_param_list:
		  declarator attr_var                    { $$ = $1; Nchappend($$, $2); }
		| arg_param_list ',' declarator attr_var {
			$$ = $1;
			Nappend($$, $3);
			Nchappend($3, $4);
		}
		;

/*
 * Declarations in beginning of blocks.
 */
block_item_list:
		  block_item                       { $$ = $1; }
		| block_item_list block_item       { $$ = Nappend($1, $2); }
		;

block_item:
		  declaration       { $$ = $1; }
		| statement         { $$ = $1; }
		;

/* Variable declaration */
declaration:
		  declaration_specifiers ';' {
				node_t *idp;

				if (($1->children != NULL) && ($1->children->next != NULL)) {
					for(idp = $1->children; idp->next->next != NULL; idp = idp->next);
				}
				else
					idp = NULL;

				if ((idp != NULL) && (idp->next->type == T_TYPECUSTOM)) {
					/* workaround for typename redeclared as variable */
					const char *id;

					id = idp->next->name;
					$$ = Nalloc(T_DECL, NULL, Nalloc(T_ID, strdup(id), NULL));
					scope_negative_typedef_add(id);
					Nfree(idp->next);
					idp->next = NULL;
					Nchinsert($$, $1);
				}
				else
					$$ = Nalloc(T_DECL, NULL, $1);
		}
		| C_TYPEDEF declaration_specifiers init_declarator_list ';' {
			 $$ = Nalloc(T_TYPEDEF, NULL, $2);
			 Nchappend($$, $3);
			 scope_typedef_add($2, $3);
			 free($1);
		}
		| declaration_specifiers init_declarator_list ';' {
			$$ = Nalloc(T_DECL, NULL, $1);
			Nappend($1, $2);
		}
		;

/* Normal declaration of variables. */
init_declarator_list:
		  init_declarator { $$ = $1; }
		| init_declarator_list ','  attr_var init_declarator_typename {
			$$ = $1;
			if ($3 != NULL) {
				Nchappend($4, $3);
				Nappend($$, $4);
			}
			else
				Nappend($$, $4);
		}
		| init_declarator_list ',' attr_var init_declarator {
			$$ = $1;
			if ($3 != NULL) {
				Nchappend($4, $3);
				Nappend($$, $4);
			}
			else
				Nappend($$, $4);
		}
		;

enum_dcl:
		  enum_head '{' moe_list optcomma '}' {
			$$ = $1;
			Nchappend($$, $3);
		}
		| C_ENUM C_NAME    { $$ = Nalloc(T_ENUM, $2, NULL); }
		;

enum_head:
		  C_ENUM             { $$ = Nalloc(T_ENUM, NULL, NULL); }
		| C_ENUM C_NAME      { $$ = Nalloc(T_ENUM, $2, NULL); }
		;

moe_list:
		  moe                { $$ = $1; }
		| moe_list ',' moe   { $$ = Nappend($1, $3); }
		;

moe:
		  C_NAME             { $$ = Nalloc(T_ID, $1, NULL); }
		| C_TYPENAME         { $$ = Nalloc(T_ID, c_typename, NULL); c_typename = NULL; Nfree($1); }
		| C_NAME '=' be      { $$ = Nalloc(T_ID, $1, Nalloc(T_INIT, NULL, $3)); }
		| C_TYPENAME '=' be  {
			$$ = Nalloc(T_ID, c_typename, Nalloc(T_INIT, NULL, $3));
			c_typename = NULL;
			Nfree($1);
		}
		;

struct_dcl:
		  str_head '{' struct_dcl_list '}' {
			$$ = $1;
			Nchappend($$, $3);
		}
		| C_STRUCT attr_var C_NAME {
			if (*$1 == 'u')
				$$ = Nalloc(T_UNION, $3, $2);
			else
				$$ = Nalloc(T_STRUCT, $3, $2);
			free($1);
		}
		| C_STRUCT attr_var C_TYPENAME {
			if (*$1 == 'u')
				$$ = Nalloc(T_UNION, c_typename, $2);
			else
				$$ = Nalloc(T_STRUCT, c_typename, $2);
			free($1);
			c_typename = NULL;
			Nfree($3);
		}
		| str_head '{' '}'    { $$ = $1; Nchappend($1, Nalloc(T_DECL, NULL, NULL)); }
		;

attr_var:
		  /* empty */        { $$ = NULL; }
		| attr_spec_list     { $$ = $1; }
		;


attr_spec_list:
		  attribute_specifier                    { $$ = $1; }
		| attr_spec_list attribute_specifier     { $$ = Nappend($1, $2); }
		;

str_head:
		  C_STRUCT attr_var {
			if (*$1 == 'u')
				$$ = Nalloc(T_UNION, NULL, $2);
			else
				$$ = Nalloc(T_STRUCT, NULL, $2);
			free($1);
		}
		| C_STRUCT attr_var str_name {
			if (*$1 == 'u')
				$$ = Nalloc(T_UNION, $3, $2);
			else
				$$ = Nalloc(T_STRUCT, $3, $2);
			free($1);
		}
		;

/* name of a structure */
str_name:
	  C_NAME        { $$ = $1; }
	| C_TYPENAME    { $$ = strdup($1->name); Nfree($1); free(c_typename); c_typename = NULL; }
	;

struct_dcl_list:
		  struct_declaration                    { if ($1->type == T_PRAGMA) $$ = $1; else $$ = Nalloc(T_DECL, NULL, $1); }
		| struct_dcl_list struct_declaration    { if ($2->type == T_PRAGMA) $$ = Nappend($1,  $2); else $$ = Nappend($1, Nalloc(T_DECL, NULL, $2)); }
		;

struct_declaration:
		  type_sq_list struct_declarator_list optsemi { $$ = Nalloc(T_TYPE, NULL, $1); Nappend($$, $2); }
		| preproc          { $$ = $1; }
		;

optsemi:
		  ';'              { }
		| optsemi ';'      { werror("extra ; in struct"); }
		;

type_sq_list:
		  type_sq                   { $$ = $1; }
		| type_sq_list type_sq      { $$ = Nappend($1, $2);  }
		;

struct_declarator_list:
		  struct_declarator                                { $$ = $1; }
		| struct_declarator_list ',' struct_declarator     { $$ = Nappend($1, $3); }
		;

struct_declarator:
		  declarator attr_var      { $$ = $1; Nchappend($$, $2); }
		| ':' be                   { $$ = Nalloc(T_ID, strdup("<anonymous>"), Nalloc(T_BITS, NULL, $2)); }
		| declarator ':' be        { $$ = $1; Nchappend($$, Nalloc(T_BITS, NULL, $3)); }
		| /* unnamed member */     { $$ = Nalloc(T_ID, strdup("<anonymous>"), NULL); }
		;

/* declarator with attributes; always preceeded by attributes */
xnfdeclarator:
		  declarator attr_var                         { $$ = $1; Nchinsert($$, $2); }
		| declarator C_ASM '(' string ')' attr_var    { $$ = $1; Nchappend($$, Nalloc(T_ASM, NULL, $4)); Nchinsert($$, $6); }
		;

xnfdeclarator_typename:
		  declarator_typename attr_var                         { $$ = $1; Nchinsert($$, $2); }
		| declarator_typename C_ASM '(' string ')' attr_var    { $$ = $1; Nchappend($$, Nalloc(T_ASM, NULL, $4)); Nchinsert($$, $6); }
		;

/*
 * Handles declarations and assignments.
 */
init_declarator:
		  xnfdeclarator                                    { $$ = $1; }
		| xnfdeclarator '=' ae                             { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, $3)); }
		| xnfdeclarator '=' '{' init_list optcomma '}'     { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, Nalloc(T_INITLIST, NULL, $4))); }
		| xnfdeclarator '=' '{' '}'                        { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, Nalloc(T_INITLIST, NULL, NULL))); }
		| xnfdeclarator '=' addrlbl                        { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, $3)); }
		;

init_declarator_typename:
		  xnfdeclarator_typename                                    { $$ = $1; }
		| xnfdeclarator_typename '=' ae                             { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, $3)); }
		| xnfdeclarator_typename '=' '{' init_list optcomma '}'     { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, Nalloc(T_INITLIST, NULL, $4))); }
		| xnfdeclarator_typename '=' '{' '}'                        { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, Nalloc(T_INITLIST, NULL, NULL))); }
		| xnfdeclarator_typename '=' addrlbl                        { $$ = $1; Nchappend($$, Nalloc(T_INIT, NULL, $3)); }
		;


initializer:
		  ae %prec ','                   { $$ = $1; }
		| addrlbl                        { $$ = $1; }
		| '{' init_list optcomma '}'     { $$ = Nalloc(T_INITLIST, NULL, $2); }
		| '{' '}'                        { $$ = Nalloc(T_INITLIST, NULL, NULL); }
		;

/* initializer content */
init_list:
		  designation initializer                   {
			if ($1 == NULL) {
				/* empty desingation is for "1", "2" in "int a[] = {1, 2, };" */
				$$ = $2;
			}
			else {
				node_t *n;
				$$ = Nalloc(T_OP, strdup("= binary"), Nappend($1, $2));
				for(n = $1; is_designator(n->children); n = n->children) ;
				Nchinsert(n, Nalloc(T_ID, strdup("<anonymous>"), NULL)); /* explicit anonymous */
			}
		}
		| init_list ','  designation initializer    {
			if ($3 != NULL) {
				node_t *n;
				/* empty desingation is for "1", "2" in "int a[] = {1, 2, };" */
				$$ = Nappend($1, Nalloc(T_OP, strdup("= binary"), Nappend($3, $4)));

				for(n = $3; is_designator(n->children); n = n->children) ;
				Nchinsert(n, Nalloc(T_ID, strdup("<anonymous>"), NULL)); /* explicit anonymous */

			}
			else
				$$ = Nappend($1, $4);
		}
		;

designation:
		  designator_list '='           { $$ = $1; }
		| '[' be C_ELLIPSIS be ']' '='  { $$ = Nop("[] binary", Nop("... binary", $2, $4), NULL); }
		|                               { $$ = NULL; }
		;

designator_list:
		  designator                   { $$ = $1; }
		| designator_list designator   { $$ = $2; Nchinsert($2, $1); }
		;

designator:
		  '[' be ']'         { $$ = Nalloc(T_OP, strdup("[] binary"), $2); }
		| C_STROP C_NAME     { $$ = Nalloc(T_OP, $1, Nalloc(T_ID, $2, NULL)); }
		;

optcomma:
		  /* VOID */
		| ','
		;

/*        STATEMENTS        */
compoundstmt:
		  beginc block_item_list endc      { $$ = $3; $$->children = $2; $$->loc.eline = loc_current->eline; $$->loc.ecol = loc_current->ecol; }
		| beginc endc                      { $$ = $2; }
		;

beginc:
		'{'      { scope_enter(); scope->compound = Nalloc(T_BLOCK, NULL, NULL); }
		;

endc:
		'}'      { $$ = scope->compound; scope_leave(); }
		;

begin:
		'{'      { scope_enter(); }
		;

end:
		'}'      { scope_leave(); }
		;

statement:
		  e ';'           { $$ = Nalloc(T_STATEMENT, strdup("expr"), $1); }
		| compoundstmt    { $$ = $1; }
		| ifprefix virt_statement {
			$$ = $1;
			Nchappend($$, Nalloc(T_THEN, NULL, $2));
		}
		| ifelprefix virt_statement {
			$$ = $1;
			Nappend($$->children, Nalloc(T_ELSE, NULL, $2));
		}
		| whprefix virt_statement                       { $$ = $1; Nchappend($$, $2); }
		| doprefix virt_statement C_WHILE '(' e ')' ';' {
			$$ = $1;
			Nchappend($$, $2);
			Nchappend($$, Nalloc(T_STATEMENT, $3, Nalloc(T_CONDITION, NULL, $5)));
		}
		| forprefix .e ')' virt_statement {
			$$ = $1;
			Nchappend($$, Nalloc(T_CONDITION, strdup("iteration"), $2));
			Nchappend($$, $4);
		}
		| switchpart statement    { $$ = $1; Nchappend($$, $2); }
		| C_BREAK  ';'            { $$ = Nalloc(T_STATEMENT, $1, NULL); }
		| C_CONTINUE  ';'         { $$ = Nalloc(T_STATEMENT, $1, NULL); }
		| C_RETURN  ';'           { $$ = Nalloc(T_STATEMENT, $1, NULL); }
		| C_RETURN e  ';'         { $$ = Nalloc(T_STATEMENT, $1, $2); }
		| C_GOTO C_NAME ';'       { $$ = Nalloc(T_STATEMENT, $1, Nalloc(T_ID, $2, NULL)); }
		| C_GOTO '*' e ';'        { $$ = Nalloc(T_STATEMENT, $1, Nop("* unary", $3, NULL));}
		| asmstatement ';'        { $$ = Nalloc(T_ASM, NULL, $1); }
		| ';'                     { $$ = NULL; }
		| error  ';'              { $$ = NULL; }
		| error '}'               { $$ = NULL; }
		| label statement         { $$ = $1; Nappend($1, $2); }
		| switch_label statement  { $$ = $1; Nappend($1, $2); }
		| preproc                 { $$ = $1; }
		;

/* wrap statement in a virtual block if it has a ->next */
virt_statement:
	statement   {
		if (($1 != NULL) && ($1->next != NULL))
			$$ = Nalloc(T_BLOCK, strdup("virtual"), $1);
		else
			$$ = $1;
		}
	;
asmstatement:
		  C_ASM asm_mvol '(' string ')'        { $$ = $4; }
		| C_ASM asm_mvol '(' string xasm ')'   { $$ = $4; Nchappend($$, $5); }
		;

asm_mvol:
		  /* empty */      { $$ = NULL; }
		| C_QUALIFIER      { $$ = Nalloc(T_QUALIFIER, $1, NULL); }
		;

xasm:
		   ':' oplist                          { $$ = Nappend(Nalloc(T_STATEMENT, strdup(":"), NULL), $2); }
		|  ':' oplist ':' oplist               { $$ = Nappend($2, Nappend(Nalloc(T_STATEMENT, strdup(":"), NULL), $4)); }
		|  ':' oplist ':' oplist ':' cnstr     { $$ = Nappend($2, Nappend(Nalloc(T_STATEMENT, strdup(":"), NULL), $4)); Nappend($$, Nappend(Nalloc(T_STATEMENT, strdup(":"), NULL), $6)); }
		;

/* asm oper list */
oplist:
		  /* nothing */    { $$ = NULL; }
		| oper             { $$ = $1; }
		;

/* asm oper */
oper:
		  string '(' e ')'             { $$ = $1; Nchappend($$, $3); }
		| oper ',' string '(' e ')'    { $$ = Nappend($1, Nappend(Nalloc(T_STATEMENT, strdup(","), NULL), $3)); Nchappend($3, $5); }
		;

/* last part of an asm statement */
cnstr:
		  string           { $$ = $1; }
		| cnstr ',' string { $$ = Nappend($1, $3); }
		;

label:
		  C_NAME ':'       { $$ = Nalloc(T_LABEL, strdup("goto"), Nalloc(T_ID, $1, NULL)); }
		| C_TYPENAME ':'   { $$ = Nalloc(T_LABEL, strdup("goto"), Nalloc(T_ID, c_typename, NULL)); c_typename = NULL; Nfree($1); }
		;

switch_label:
		  C_CASE be ':'                 { $$ = Nalloc(T_LABEL, $1, $2); }
		| C_CASE be C_ELLIPSIS be ':'   { $$ = Nalloc(T_LABEL, $1, Nop("... binary", $2, $4)); }
		| C_DEFAULT ':'                 { $$ = Nalloc(T_LABEL, $1, NULL); }
		;

doprefix:
		C_DO { $$ = Nalloc(T_STATEMENT, $1, NULL); }
		;

ifprefix:
		C_IF '(' e ')' { $$ = Nalloc(T_STATEMENT, $1, Nalloc(T_CONDITION, NULL, $3)); }
		;

ifelprefix:
		ifprefix statement C_ELSE {
			$$ = $1;
			Nchappend($$, Nalloc(T_THEN, NULL, $2));
		}
		;

whprefix:
		C_WHILE  '('  e  ')'    { $$ = Nalloc(T_STATEMENT, $1, Nalloc(T_CONDITION, NULL, $3)); }
		;
forprefix:
		  C_FOR  '('  .e  ';' .e  ';' {
			$$ = Nalloc(T_STATEMENT, $1, NULL);
			Nchappend($$, Nalloc(T_CONDITION, strdup("init"), $3));
			Nchappend($$, Nalloc(T_CONDITION, strdup("condition"), $5));
		}
		| C_FOR '(' {} declaration .e ';' {
			$$ = Nalloc(T_STATEMENT, $1, NULL);
			Nchappend($$, Nalloc(T_CONDITION, strdup("init"), $4));
			Nchappend($$, Nalloc(T_CONDITION, strdup("condition"), $5));
		}
		;

switchpart:
		C_SWITCH  '('  e ')' { $$ = Nalloc(T_STATEMENT, $1, Nalloc(T_CONDITION, NULL, $3)); }
		;


/*        EXPRESSIONS        */

/* expression or empty */
.e:
		  e    { $$ = $1; }
		|      { $$ = NULL; }
		;

/* list of expressions */
elist:
		  /* empty */            { $$ = NULL; }
		| ae %prec ','           { $$ = Nalloc(T_ARG, NULL, $1); }
		| elist  ','  ae         { $$ = Nappend($1, Nalloc(T_ARG, NULL, $3)); }
		| elist  ','  cast_type  { /* hack for stdarg */ $$ = Nappend($1, Nalloc(T_ARG, NULL, $3)); }
		;

/*
 * expression; precedence order of operators.
 */
e:
		  e ',' ae         {
			if ($1->type == T_EXPRLIST) {
				$$ = $1;
				Nchappend($$, $3);
			}
			else
				$$ = Nalloc(T_EXPRLIST, NULL, Nappend($1, $3));
		}
		| ae               { $$ = $1; }
		;
/* assignment-expression, it cannot be a comma separated list of expressions */
ae:
		  term '=' ae      { $$ = Nop("= binary", $1, $3); }
		| term '=' addrlbl { $$ = Nop("= binary", $1, $3); }
		| be               { $$ = $1; }
		| term C_ASOP ae   {
				$2 = realloc($2, 64);
				strcat($2, " binary");
				$$ = Nop($2, $1, $3);
				free($2);
			}

		;
/* binary-, ternary- or unary-expression, it cannot be an assignment */
be:
		  be '?' e ':' be  { $$ = Nop("? trenary", $1, $3); Nchappend($$, $5); }
		| be '?' ':' be    { $$ = Nop("? trenary", $1, NULL); Nchappend($$, $4); }
		| be C_OROR be     { $2 = realloc($2, 64); strcat($2, " binary"); $$ = Nop($2, $1, $3); free($2); }
		| be C_ANDAND be   { $2 = realloc($2, 64); strcat($2, " binary"); $$ = Nop($2, $1, $3); free($2); }
		| be '|' be        { $$ = Nop("| binary", $1, $3); }
		| be '^' be        { $$ = Nop("^ binary", $1, $3); }
		| be '&' be        { $$ = Nop("& binary", $1, $3); }
		| be C_EQUOP be    { $2 = realloc($2, 64); strcat($2, " binary"); $$ = Nop($2, $1, $3); free($2); }
		| be C_RELOP be    { $2 = realloc($2, 64); strcat($2, " binary"); $$ = Nop($2, $1, $3); free($2); }
		| be C_SHIFTOP be  { $2 = realloc($2, 64); strcat($2, " binary"); $$ = Nop($2, $1, $3); free($2); }
		| be '+' be        { $$ = Nop("+ binary", $1, $3); }
		| be '-' be        { $$ = Nop("- binary", $1, $3); }
		| be C_DIVOP be    { $2 = realloc($2, 64); strcat($2, " binary"); $$ = Nop($2, $1, $3); free($2); }
		| be '*' be        { $$ = Nop("* binary", $1, $3); }
		| term             { $$ = $1; }
		;

/* jump address pointer, for gcc && goto trick */
addrlbl:
		C_ANDAND C_NAME    { $$ = Nop("&& unary", Nalloc(T_ID, $2, NULL), NULL); free($1); }
		;

term:
		  term C_INCOP       { char *s; s = strdup("post?? unary"); memcpy(s+4, $2, 2); $$ = Nop(s, $1, NULL); free($2); free(s); }
		| '*' term           { $$ = Nop("* unary", $2, NULL); }
		| '&' term           { $$ = Nop("& unary", $2, NULL); }
		| '-' term           { $$ = Nop("- unary", $2, NULL); }
		| '+' term           { $$ = Nop("+ unary", $2, NULL); }
		| C_UNOP term        { $$ = Nop($1, $2, NULL); free($1); }
		| C_INCOP term       { char *s; s = strdup("pre?? unary"); memcpy(s+3, $1, 2); $$ = Nop(s, $2, NULL); free($1); free(s); }
		| C_SIZEOF term      { $$ = Nop("sizeof unary", $2, NULL); }
		| '(' cast_type ')' term  %prec C_INCOP              { $$ = Nop("cast binary", Nalloc(T_TYPE, NULL, $2), $4); }
		| '(' cast_type ')' '{' init_list optcomma '}'       { $$ = Nop("cast binary", Nalloc(T_TYPE, NULL, $2), Nalloc(T_INITLIST, NULL, $5)); }
		| C_SIZEOF '(' cast_type ')'  %prec C_SIZEOF         { $$ = Nop("sizeof unary", Nalloc(T_TYPE, NULL, $3), NULL); }
		| C_ALIGNOF '(' cast_type ')'                        { $$ = Nop("alignof unary", $3, NULL); }
		| term '[' e ']'                        { $$ = Nalloc(T_OP, strdup("[] binary"), $1); Nchappend($$, $3); }
		| C_NAME  '(' elist ')'                 { $$ = Nalloc(T_CALL, $1, Nappend(Nalloc(T_ID, strdup($1), NULL), Nalloc(T_ARGLIST, NULL, $3))); }
		| term  '(' elist ')'                   { $$ = Nalloc(T_CALL, strdup(extract_id_from_cast($1)), $1); Nchappend($$, Nalloc(T_ARGLIST, NULL, $3)); }
		| term C_STROP C_NAME                   { $$ = Nop($2, $1, Nalloc(T_ID, $3, NULL)); free($2); }
		| term C_STROP C_TYPENAME               { $$ = Nop($2, $1, Nalloc(T_ID, c_typename, NULL)); c_typename = NULL; free($2); Nfree($3); }
		| C_NAME %prec C_SIZEOF                 { $$ = Nalloc(T_ID, $1, NULL); }
		| C_ICON                                { $$ = Nalloc(T_ICONST, $1, NULL); }
		| C_FCON                                { $$ = Nalloc(T_FCONST, $1, NULL); }
		| string                                { $$ = $1; }
		| '('  e  ')'                           { $$ = $2; }
		| '(' begin block_item_list end ')'     { $$ = Nalloc(T_BLOCK, NULL, $3); }
		;

/* a single string or multiple strings appended */
string_:
		  C_STRING             { $$ = Nalloc(T_SCONST, $1, NULL); }
		| string_ C_STRING     { $$ = Nappend($1, Nalloc(T_SCONST, $2, NULL)); }
		;

/* if we have multiple strings appended, wrap the list in SCONSTLIST */
string:
	string_                  { if ($1->next != NULL) $$ = Nalloc(T_SCONSTLIST, NULL, $1); else $$ = $1; }
	;

/* the inner part of casts, list of types */
cast_type:
		  type_sq_list                        { $$ = $1; }
		| type_sq_list abstract_declarator    { $$ = Nappend($1, $2); }
		;

/* things that shouldn't have hit this stage:
   preprocessor directives, comments */
preproc:
	  C_PRAGMA              { $$ = Nalloc(T_PRAGMA, $1, NULL); }
	| C_COMMENT             { $$ = Nalloc(T_COMMENT, $1, NULL); }
	;

%%

static const char *extract_id_from_cast(node_t *n)
{
	switch(n->type) {
		case T_OP:
			if (strncmp(n->name, "cast ", 5) == 0)
				return n->children->next->name;
			if ((strcmp(n->name, "* unary") == 0) || (strcmp(n->name, "[] binary") == 0) ||
			    (strcmp(n->name, "-> binary") == 0) || (strcmp(n->name, ". binary") == 0))
				return "<anonymous>";
			Nprint(" ? ", n);
			fprintf(stderr, "c99tree error: strange operator on function - please include that line in your bugreport!\n");
			abort();
		case T_ID:
			return "<anonymous>";
		default:
			Nprint(" ? ", n);
			fprintf(stderr, "c99tree error: strange function cast - please include that line in your bugreport!\n");
			abort();
	}
}

static int is_designator(node_t *n)
{
	return (n->type == T_OP) && ((strcmp(n->name, "[] binary") == 0) || (strcmp(n->name, ". binary") == 0) || (strcmp(n->name, "-> binary") == 0));
}