typedef struct loc_file_s {
	char *fn;              /* file name, as present in #include */
	char *path;            /* full path (note search paths) */
} loc_file_t;

typedef struct loc_s {
	loc_file_t *file;      /* host file */
	int soffs, eoffs;      /* byte offsets from the start of the file (start/end) counting from 0 */
	int sline, scol;       /* start line/col counting from 1 */
	int eline, ecol;       /* end line/col counting from 1 */
} loc_t;

/* current location we are reading tokens at (top of the location stack) */
extern loc_t *loc_current;

/* format string for printing location */
extern char *loc_format;

/* print location using loc_format in a static buffer */
const char *loc_print(const node_t *node);

/* convert a location found in in, using loc_format in a static buffer (returned) */
loc_t *loc_scan(const char *in);

/* Open a new file, set it as current loc_file, pushing the old current on
   top of the stack; hash files by path */
void loc_file_open(const char *fn, const char *path);

/* Close current loc_file, pop loc from the stack, make that the current one */
void loc_file_close(void);

/* Copy current loc end fields to start fields, advance the end
   fields as needed using token string; returns token to shorten source
   of the .l */
const char *loc_update_end(const char *token);

/* Copy location struct */
void loc_cpy(loc_t *dst, const loc_t *src);

/* Delete the location files database and stack */
void loc_destroy();
