%{
/*
 * Copyright (c) 2002 Anders Magnusson. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

 /* This is a modified verison for c99tree; these modifications are
    Copyright (c) 2010 Tibor Palinkas.
    Original file version marking:
    scan.l,v 1.94 2010/08/11 14:08:44 ragge Exp */
#undef YY_NO_INPUT
%}


D     [0-9]
L     [a-zA-Z_]
H     [a-fA-F0-9]
E     [Ee][+-]?{D}+
P     [Pp][+-]?{D}+
FS    (f|F|l|L)?i?
IS    (u|U|l|L)*
UL    ({L}|\\u{H}{H}{H}{H}|\\U{H}{H}{H}{H}{H}{H}{H}{H})
NL    [\\][\n]
ANY   (.|{NL})

%{
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include "tree.h"
#include "scope.h"
#include "cgram.h"
#include "c99tree.h"

static int is_typename();
static void control(int);
static void pragma(void);

#define CPP_IDENT 2
#define CPP_LINE 3
#define CPP_HASH 4

#if defined(FLEX_SCANNER) && YY_FLEX_SUBMINOR_VERSION == 31
/* Hack to avoid unnecessary warnings */
FILE *yyget_in  (void);
FILE *yyget_out  (void);
int yyget_leng  (void);
char *yyget_text  (void);
void yyset_in (FILE *);
void yyset_out (FILE *);
int yyget_debug  (void);
void yyset_debug (int);
int yylex_destroy  (void);
extern int yyget_lineno (void);
extern void yyset_lineno (int);
#endif

/* shorthand for loc_update_end for keeping track of location */
static void lue()
{
	loc_update_end(yytext);
}

/* same as lue, but also strdups yytext into yylval.string */
static void luedup()
{
	yylval.string = strdup(loc_update_end(yytext));
}

void cunput(char c);
static int read_comment();
%}

%%
"/*" { return read_comment(); }

"//"(.|{NL})*       {
		lue();
		yylval.string = strdup(yytext+2);

		return(C_COMMENT);
	}

"__func__"       { luedup(); return(C_STRING); }
"__attribute__"  { lue();    return(C_ATTRIBUTE); }
"asm"            { lue();    return(C_ASM); }
"auto"           { luedup(); return(C_CLASS); }
"_Bool"          { luedup(); return (C_TYPE); }
"break"          { luedup(); return(C_BREAK); }
"case"           { luedup(); return(C_CASE); }
"char"           { luedup(); return (C_TYPE); }
"_Complex"       { luedup(); return (C_TYPESPEC); }
"const"          { luedup(); return (C_QUALIFIER); }
"continue"       { luedup(); return(C_CONTINUE); }
"default"        { luedup(); return(C_DEFAULT); }
"do"             { luedup(); return(C_DO); }
"double"         { luedup(); return (C_TYPE); }
"else"           { lue();    return(C_ELSE); }
"enum"           { lue();    return(C_ENUM); }
"extern"         { luedup(); return (C_CLASS); }
"float"          { luedup(); return (C_TYPE); }
"for"            { luedup(); return(C_FOR); }
"goto"           { luedup(); return(C_GOTO); }
"if"             { luedup(); return(C_IF); }
"_Imaginary"     { luedup(); return (C_TYPESPEC); }
"inline"         { luedup(); return(C_INLINE); }
"int"            { luedup(); return (C_TYPE); }
"long"           { luedup(); return (C_TYPESPEC); }
"register"       { luedup(); return (C_CLASS); }
"restrict"       { luedup(); return (C_QUALIFIER); }
"return"         { luedup(); return(C_RETURN); }
"short"          { luedup(); return (C_TYPESPEC); }
"signed"         { luedup(); return (C_TYPESPEC); }
"sizeof"         { lue();    return(C_SIZEOF); }
"typeof"         { lue();    return(C_TYPEOF); }
"static"         { luedup(); return (C_CLASS); }
"struct"         { luedup(); return(C_STRUCT); }
"switch"         { luedup(); return(C_SWITCH); }
"typedef"        { luedup(); return (C_TYPEDEF); }
"union"          { luedup(); return(C_STRUCT); }
"unsigned"       { luedup(); return (C_TYPESPEC); }
"void"           { luedup(); return (C_TYPE); }
"volatile"       { luedup(); return (C_QUALIFIER); }
"while"          { luedup(); return(C_WHILE); }

{UL}({UL}|{D})*  { luedup(); return is_typename();}

0[xX]{H}+{IS}?   { luedup(); return(C_ICON); }
0{D}+{IS}?       { luedup(); return(C_ICON); }
{D}+{IS}?        { luedup(); return(C_ICON); }
L'(\\.|[^\\'])+' { luedup(); return(C_ICON); }
'(\\.|[^\\'])+'  { luedup(); return(C_ICON); }

{D}+{E}{FS}?             { luedup(); return(C_FCON); }
{D}*"."{D}+({E})?{FS}?   { luedup(); return(C_FCON); }
{D}+"."{D}*({E})?{FS}?   { luedup(); return(C_FCON); }
0[xX]{H}*"."{H}+{P}{FS}? { luedup(); return(C_FCON); }
0[xX]{H}+"."{P}{FS}?     { luedup(); return(C_FCON); }
0[xX]{H}+{P}{FS}?        { luedup(); return(C_FCON); }

L?\"(\\.|[^\\"]|{NL})*\"      { luedup(); return C_STRING; }

"..."         { lue(); return(C_ELLIPSIS); }
">>="         { luedup(); return(C_ASOP); }
"<<="         { luedup(); return(C_ASOP); }
"+="          { luedup(); return(C_ASOP); }
"-="          { luedup(); return(C_ASOP); }
"*="          { luedup(); return(C_ASOP); }
"/="          { luedup(); return(C_ASOP); }
"%="          { luedup(); return(C_ASOP); }
"&="          { luedup(); return(C_ASOP); }
"^="          { luedup(); return(C_ASOP); }
"|="          { luedup(); return(C_ASOP); }
">>"          { luedup(); return(C_SHIFTOP); }
"<<"          { luedup(); return(C_SHIFTOP); }
"++"          { luedup(); return(C_INCOP); }
"--"          { luedup(); return(C_INCOP); }
"&&"          { luedup(); return(C_ANDAND); }
"||"          { luedup(); return(C_OROR); }
"<="          { luedup(); return(C_RELOP); }
">="          { luedup(); return(C_RELOP); }
"=="          { luedup(); return(C_EQUOP); }
"!="          { luedup(); return(C_EQUOP); }
";"           { lue();    return(';'); }
("{"|"<%")    { lue();    return('{'); }
("}"|"%>")    { lue();    return('}'); }
","           { lue();    return(','); }
":"           { lue();    return(':'); }
"="           { lue();    return('='); }
"("           { lue();    return('('); }
")"           { lue();    return(')'); }
("["|"<:")    { lue();    return('['); }
("]"|":>")    { lue();    return(']'); }
"&"           { lue();    return('&'); }
"->"          { lue(); yylval.string = strdup("-> binary");  return(C_STROP); }
"."           { lue(); yylval.string = strdup(". binary");  return(C_STROP); }
"!"           { lue(); yylval.string = strdup("! unary"); return(C_UNOP); }
"~"           { lue(); yylval.string = strdup("~ unary"); return(C_UNOP); }
"-"           { lue();    return('-'); }
"+"           { lue();    return('+'); }
"*"           { lue();    return('*'); }
"/"           { luedup(); return(C_DIVOP); }
"%"           { luedup(); return(C_DIVOP); }
"<"           { luedup(); return(C_RELOP); }
">"           { luedup(); return(C_RELOP); }
"^"           { lue();    return('^'); }
"|"           { lue();    return('|'); }
"?"           { lue();    return('?'); }

^#pragma[ \t]{ANY}*  { pragma(); return C_PRAGMA; }
^#ident[ \t]{ANY}*   { control(CPP_IDENT); }
^#line[ \t]{ANY}*    { control(CPP_LINE); }
^#{ANY}*             { control(CPP_HASH); }

[ \t\v\f\r]      { lue(); }
"\n"             { lue(); }
.                { lue(); fprintf(stderr, "Warning: lex ignoring character '%c'\n", *yytext); /* ignore bad characters */ }

%%

int
yywrap(void)
{
	if (0) unput(0); /* quiet gcc */
	return(1);
}


void
control(int t)
{
	char *wr = yytext;
	char *eptr;
	int val;

	lue();

	wr++;             /* Skip initial '#' */
	switch (t) {
	case CPP_IDENT:
		return;         /* Just skip these for now. */

	case CPP_LINE:
		wr += 4;
		/* FALLTHROUGH */
	case CPP_HASH:
		val = strtol(wr, &eptr, 10);
		if (wr == eptr) /* Illegal string */
			goto bad;
		wr = eptr;
//		lineno = val - 1;
		while (*wr && *wr != '\"')
			wr++;
		if (*wr == 0)
			return;
		if (*wr++ != '\"')
			goto bad;
		eptr = wr;
		while (*wr && *wr != '\"')
			wr++;
		if (*wr != '\"')
			goto bad;
		*wr = 0;
/*		ftitle = strdup(eptr); */
	}
	return;
bad:
	fprintf(stderr, "%s: illegal control", yytext);
}

/*
 * got a full pragma line.  Split it up here.
 */
static void pragma()
{
	char *arg;

	lue();
	arg = yytext+7;

	while((*arg == ' ') || (*arg == '\t')) arg++;
	yylval.string = strdup(arg);
}

void cunput(char c)
{
	unput(c);
}


extern char *c_typename;
static int is_typename()
{
	const node_t *n;
	n = scope_get_typedef(yylval.string);
	if (n != NULL) {
		if (c_typename != NULL)
			free(c_typename);
		c_typename = yylval.string;
		yylval.node = Nclone(n);
		return C_TYPENAME;
	}
	return C_NAME;
}

static char *comment = NULL;
static int alloced = 0, used = 0;

static void cappend(char c)
{
	if (used == alloced) {
		alloced += 256;
		comment = realloc(comment, alloced);
	}
	comment[used] = c;
	used++;
}


/* read in comment as documented in flex doc */
static int read_comment()
{
	int c;
  for (;;) {
		while(((c = input()) != '*') && (c != EOF)) cappend(c);    /* eat up text of comment */
    if (c == '*') {
			cappend(c);
			while ((c = input()) == '*') cappend(c);
	    if ( c == '/' ) {
				comment[used-1] = '\0';
				yylval.string = comment;
				comment = 0;
				alloced = 0;
				used = 0;
				return C_COMMENT;
			}
			else
				cappend(c);
	  }
    if (c == EOF) {
			fprintf(stderr, "EOF in /* */ comment");
			break;
		}
	}
	/* we never get here */
	return -1;
}
